==> FHLBBS-4.2.tgz: maple/io.c <==
/*-------------------------------------------------------*/
/* io.c         ( NTHU CS MapleBBS Ver 2.36 )            */
/*-------------------------------------------------------*/
/* target : basic console/screen/keyboard I/O routines   */
/* create : 95/02/28                                     */
/* update : 95/12/15                                     */
/*-------------------------------------------------------*/

#include "bbs.h"

#ifdef AIX
#include <sys/select.h>
#endif

#ifdef  Linux
#define OBUFSIZE  (2048)
#define IBUFSIZE  (128)
#else
#define OBUFSIZE  (4096)
#define IBUFSIZE  (256)
#endif

#define INPUT_ACTIVE    0
#define INPUT_IDLE      1

static char outbuf[OBUFSIZE];
static int obufsize = 0;

static char inbuf[IBUFSIZE];
static int ibufsize = 0;
static int icurrchar = 0;

static int i_mode = INPUT_ACTIVE;

extern int dumb_term;


/* ----------------------------------------------------- */
/* wܰʺAݪO                                      */
/* ----------------------------------------------------- */


#ifdef  HAVE_MOVIE
#define STAY_TIMEOUT    (30*60)
extern void movie();

static void
hit_alarm_clock()
{
  static int stay_time = 0;
  static int idle_time = 0;

  if (currutmp->pid != currpid)
  {
    setup_utmp(XMODE);          /* stm shm */
  }

  if ((i_mode == INPUT_IDLE) && (!HAS_PERM(PERM_NOTIMEOUT)))
  {
    idle_time += MOVIE_INT;
    if (idle_time > IDLE_TIMEOUT)
    {
      clear();
      fprintf(stderr, "WLmɶIBooting...\n");
      abort_bbs();
    }
  }
  else
  {
    idle_time = 0;
    i_mode = INPUT_IDLE;
  }

  if (currstat && currstat < CLASS)
    movie(0);
  alarm(MOVIE_INT);
 

  stay_time += MOVIE_INT;
  if (chkmail(0) && stay_time > 10 * 60 || stay_time > STAY_TIMEOUT)
  {
    /* bo̴ user 𮧤@U */
    char *msg[8] = {"y, ||, ܤf....",
                    "zwgFbӤpo! _Ӭʤ@Ua!",
                    "HR±zR",
                    "AUh||[`O~~~",
                    "SnƱѤFBzO?",
                    "ɥiHQQSHݭnߡH",
		    "OѤFnWWZҡAܳܤAHKoH֪εhI",
                    "OѤFq᭱]OӬͥͪHI"};
/*
woju
*/
    time_t now = time(0);

    char buf[100];
    int i = rand() % 8;

    sprintf(currmsg, "[1;33;41m[%s] %s[m", Cdate(&now), 
       chkmail(0) ? "Hc٦SݹLH@!" : msg[i]);
    kill(currpid, SIGUSR2);
    stay_time = 0;
  }
  if (idle_time > IDLE_TIMEOUT - 60)
  {
    sprintf(currmsg, "[1;5;37;41mĵiGzwmL[A"
                       "YL^AtΧYNII[m");
    kill(currpid, SIGUSR2);
  }
}


void
init_alarm()
{
  alarm(0);
  signal(SIGALRM, hit_alarm_clock);
  alarm(MOVIE_INT);
}

#else                           /* HAVE_MOVIE */

static void
hit_alarm_clock()
{
  if (HAS_PERM(PERM_NOTIMEOUT))
    return;
  if (i_mode == INPUT_IDLE)
  {
    clear();
    fprintf(stderr, "WLmɶIBooting...\n");
    kill(currpid, SIGHUP);
  }
  i_mode = INPUT_IDLE;
  alarm(IDLE_TIMEOUT);
}


void
init_alarm()
{
  signal(SIGALRM, hit_alarm_clock);
  alarm(IDLE_TIMEOUT);
}
#endif                          /* HAVE_MOVIE */


/* ----------------------------------------------------- */
/* output routines                                       */
/* ----------------------------------------------------- */


void
oflush()
{
  if (obufsize)
  {
    write(1, outbuf, obufsize);
    obufsize = 0;
  }
}


void
output(s, len)
  char *s;
{
  /* Invalid if len >= OBUFSIZE */

  if (obufsize + len > OBUFSIZE)
  {
    write(1, outbuf, obufsize);
    obufsize = 0;
  }
  memcpy(outbuf + obufsize, s, len);
  obufsize += len;
}


void
ochar(c)
{
  if (obufsize > OBUFSIZE - 1)
  {
    write(1, outbuf, obufsize);
    obufsize = 0;
  }
  outbuf[obufsize++] = c;
}


/* ----------------------------------------------------- */
/* input routines                                        */
/* ----------------------------------------------------- */


static int i_newfd = 0;
static struct timeval i_to, *i_top = NULL;
static int (*flushf) () = NULL;


void
add_io(fd, timeout)
  int fd;
  int timeout;
{
  i_newfd = fd;
  if (timeout)
  {
    i_to.tv_sec = timeout;
    i_to.tv_usec = 0;
    i_top = &i_to;
  }
  else
    i_top = NULL;
}


void
add_flush(flushfunc)
  int (*flushfunc) ();
{
  flushf = flushfunc;
}


int
num_in_buf()
{
  return icurrchar - ibufsize;
}


int
igetch()
{
  int ch;

  for (;;)
  {
    if (ibufsize == icurrchar)
    {
      fd_set readfds;
      struct timeval to;

      to.tv_sec = to.tv_usec = 0;
      FD_ZERO(&readfds);
      FD_SET(0, &readfds);
      if (i_newfd)
        FD_SET(i_newfd, &readfds);
      if ((ch = select(FD_SETSIZE, &readfds, NULL, NULL, &to)) <= 0)
      {
        if (flushf)
          (*flushf) ();

        if (dumb_term)
          oflush();
        else
          refresh();

        FD_ZERO(&readfds);
        FD_SET(0, &readfds);
        if (i_newfd)
          FD_SET(i_newfd, &readfds);

        while ((ch = select(FD_SETSIZE, &readfds, NULL, NULL, i_top)) < 0)
        {
          if (errno == EINTR)
            continue;
          else
          {
            perror("select");
            return -1;
          }
        }
        if (ch == 0)
          return I_TIMEOUT;
      }
      if (i_newfd && FD_ISSET(i_newfd, &readfds))
        return I_OTHERDATA;

      while ((ibufsize = read(0, inbuf, IBUFSIZE)) <= 0)
      {
        if (ibufsize == 0)
          longjmp(byebye, -1);
        if (ibufsize < 0 && errno != EINTR)
          longjmp(byebye, -1);
      }
      icurrchar = 0;
    }

    i_mode = INPUT_ACTIVE;
    ch = inbuf[icurrchar++];
    if (ch != Ctrl('L'))
      return (ch);

    redoscr();
  }
}

char* phone_char(char c)
{
   switch (c) {
          case '1':
             return "t";
          case 'q':
             return "u";
          case 'a':
             return "v";
          case 'z':
             return "w";
          case '2':
             return "x";
          case 'w':
             return "y";
          case 's':
             return "z";
          case 'x':
             return "{";
          case 'e':
             return "|";
          case 'd':
             return "}";
          case 'c':
             return "~";
          case 'r':
             return "";
          case 'f':
             return "";
          case 'v':
             return "";
          case '5':
             return "";
          case 't':
             return "";
          case 'g':
             return "";
          case 'b':
             return "";
          case 'y':
             return "";
          case 'h':
             return "";
          case 'n':
             return "";
          case 'u':
             return "";
          case 'j':
             return "";
          case 'm':
             return "";
          case '8':
             return "";
             break;
          case 'i':
             return "";
          case 'k':
             return "";
          case ',':
             return "";
          case '9':
             return "";
          case 'o':
             return "";
          case 'l':
             return "";
          case '.':
             return "";
          case '0':
             return "";
          case 'p':
             return "";
          case ';':
             return "";
          case '/':
             return "";
          case '-':
             return "";
          case '6':
             return "";
          case '3':
             return "";
          case '4':
             return "";
          case '7':
             return "";
   }
  return 0;
}


int
getdata(line, col, prompt, buf, len, echo)
  int line, col;
  char *prompt, *buf;
  int len, echo;
{
  register int ch;
  int clen;
  int x, y;
  extern unsigned char scr_cols;
#define MAXLASTCMD 6
  static char lastcmd[MAXLASTCMD][80];

  if (prompt)
  {
    move(line, col);
    clrtoeol();
    outs(prompt);
  }
  else
    clrtoeol();

  if (dumb_term || !echo)
  {
    len--;
    clen = 0;
    while ((ch = igetch()) != '\r')
    {
      if (ch == '\n')
        break;
      if (ch == '\177' || ch == Ctrl('H'))
      {
        if (!clen)
        {
          bell();
          continue;
        }
        clen--;
        if (echo)
        {
          ochar(Ctrl('H'));
          ochar(' ');
          ochar(Ctrl('H'));
        }
        continue;
      }

#ifdef BIT8
      if (!isprint2(ch))
#else
      if (!isprint(ch))
#endif

      {
        if (echo)
          bell();
        continue;
      }
      if (clen >= len)
      {
        if (echo)
          bell();
        continue;
      }
      buf[clen++] = ch;
      if (echo)
        ochar(ch);
    }
    buf[clen] = '\0';
    outc('\n');
    oflush();
  }
  else
  {
   int cmdpos = -1;
   int currchar = 0;

    getyx(&y, &x);
    standout();
    for (clen = len--; clen; clen--)
      outc(' ');
    standend();
    memset(buf, 0, len);

    while (move(y, x + currchar), (ch = igetkey()) != '\r')
    {
/*
woju
*/
       switch (ch) {

       case KEY_DOWN:
       case Ctrl('N'):
          cmdpos += MAXLASTCMD - 2;
       case Ctrl('P'):
       case KEY_UP: {
          int i;

          cmdpos++;
          cmdpos %= MAXLASTCMD;
          strncpy(buf, lastcmd[cmdpos], len);

          move(y, x);                   /* clrtoeof */
          for (i = 0; i <= clen; i++)
             outc(' ');
          move(y, x);

          outs(buf);
          clen = currchar = strlen(buf);
          continue;
       }
       case KEY_LEFT:
          if (currchar)
             --currchar;
          continue;
       case KEY_RIGHT:
          if (buf[currchar])
             ++currchar;
          continue;
       }

      if (ch == '\n' || ch == '\r')
         break;
/*
woju
*/
      if (currutmp->msgs[0].last_pid && ch == Ctrl('R') &&
          !(currutmp->mode == 0 &&
            (currutmp->chatid[0] == 2 || currutmp->chatid[0] == 3))) {
         extern screenline* big_picture;
         screenline* screen0 = calloc(t_lines, sizeof(screenline));

         memcpy(screen0, big_picture, t_lines * sizeof(screenline));

         show_last_call_in();
         my_write(currutmp->msgs[0].last_pid, "Te^hG");
         memcpy(big_picture, screen0, t_lines * sizeof(screenline));
         free(screen0);
         redoscr();
         continue;
      }                                                                       
      if (ch == Ctrl('U') && currstat != IDLE  &&
          !(currutmp->mode == 0 &&
            (currutmp->chatid[0] == 2 || currutmp->chatid[0] == 3))) {
         extern screenline* big_picture;
         screenline* screen0 = calloc(t_lines, sizeof(screenline));

         memcpy(screen0, big_picture, t_lines * sizeof(screenline));

         t_users();
         memcpy(big_picture, screen0, t_lines * sizeof(screenline));
         free(screen0);
         redoscr();
         continue;
      }                                                                       
      if (ch == Ctrl('I') && currstat != IDLE &&
          !(currutmp->mode == 0 &&
            (currutmp->chatid[0] == 2 || currutmp->chatid[0] == 3))) {
         extern screenline* big_picture;
         screenline* screen0 = calloc(t_lines, sizeof(screenline));

         memcpy(screen0, big_picture, t_lines * sizeof(screenline));

         t_idle();
         memcpy(big_picture, screen0, t_lines * sizeof(screenline));
         free(screen0);
         redoscr();
         continue;
      }                                                                       
      if (ch == '\177' || ch == Ctrl('H'))
      {
        if (currchar) {
           int i;

           currchar--;
           clen--;
           for (i = currchar; i <= clen; i++)
              buf[i] = buf[i + 1];
           move(y, x + clen);
           outc(' ');
           move(y, x);
           outs(buf);
        }
/*
        if (clen)
        {
          clen--;
          move(y, x + clen);
          outc(' ');
          move(y, x + clen);
        }
*/
        continue;
      }
      if (ch == Ctrl('Y')) {
         int i;

         buf[0] = 0;
         currchar = 0;
         move(y, x);                    /* clrtoeof */
         for (i = 0; i < clen; i++)
            outc(' ');
         clen = 0;
         continue;
      }
      if (ch == Ctrl('C')) {
         buf[0] = 0;
         break;
      }
      if (ch == Ctrl('D')) {
        if (buf[currchar]) {
           int i;

           clen--;
           for (i = currchar; i <= clen; i++)
              buf[i] = buf[i + 1];
           move(y, x + clen);
           outc(' ');
           move(y, x);
           outs(buf);
        }
        continue;
      }
      if (ch == Ctrl('K')) {
         int i;

         buf[currchar] = 0;
         move(y, x + currchar);
         for (i = currchar; i < clen; i++)
            outc(' ');
         clen = currchar;
         continue;
      }
      if (ch == Ctrl('A')) {
         currchar = 0;
         continue;
      }
      if (ch == Ctrl('E')) {
         currchar = clen;
         continue;
      }


#ifdef BIT8
      if (!isprint2(ch))
#else
      if (!isprint(ch))
#endif

      {
        continue;
      }
      if (clen >= len || x + clen >= scr_cols)
      {
        continue;
      }
/*
woju
*/
      if (buf[currchar]) {               /* insert */
         int i;

         for (i = currchar; buf[i] && i < len && i < 80; i++)
            ;
         buf[i+1] = 0;
         for (; i > currchar; i--)
            buf[i] = buf[i - 1];
      }
      else                              /* append */
         buf[currchar + 1] = '\0';
      buf[currchar] = ch;
      move(y, x + currchar);
      outs(buf + currchar);
      currchar++;
      clen++;
/*
      move(y, x + clen);
      buf[clen++] = ch;
      outc(echo ? ch : '*');
*/
    }
    buf[clen] = '\0';
    if (clen > 1) {
       for (cmdpos = MAXLASTCMD - 1; cmdpos; cmdpos--)
          strcpy(lastcmd[cmdpos], lastcmd[cmdpos - 1]);
       strncpy(lastcmd[0], buf, len);
    }
    if (echo)
      outc('\n');
    refresh();
  }
  if ((echo == LCECHO) && ((ch = buf[0]) >= 'A') && (ch <= 'Z'))
    buf[0] = ch | 32;
  return clen;
}


/*
woju
*/
#define TRAP_ESC

#ifdef  TRAP_ESC
int KEY_ESC_arg;

int
igetkey()
{
  int mode;
  int ch, last;

  mode = last = 0;
  while (1)
  {
    ch = igetch();
    if (mode == 0)
    {
      if (ch == KEY_ESC)
        mode = 1;
      else
        return ch;              /* Normal Key */
    }
    else if (mode == 1)
    {                           /* Escape sequence */
      if (ch == '[' || ch == 'O')
        mode = 2;
      else if (ch == '1' || ch == '4')
        mode = 3;
      else
      {
        KEY_ESC_arg = ch;
        return KEY_ESC;
      }
    }
    else if (mode == 2)
    {                           /* Cursor key */
      if (ch >= 'A' && ch <= 'D')
        return KEY_UP + (ch - 'A');
      else if (ch >= '1' && ch <= '6')
        mode = 3;
      else
        return ch;
    }
    else if (mode == 3)
    {                           /* Ins Del Home End PgUp PgDn */
      if (ch == '~')
        return KEY_HOME + (last - '1');
      else
        return ch;
    }
    last = ch;
  }
}

#else                           /* TRAP_ESC */

int
igetkey(void)
{
  int mode;
  int ch, last;

  mode = last = 0;
  while (1)
  {
    ch = igetch();
    if (ch == KEY_ESC)
      mode = 1;
    else if (mode == 0)         /* Normal Key */
      return ch;
    else if (mode == 1)
    {                           /* Escape sequence */
      if (ch == '[' || ch == 'O')
        mode = 2;
      else if (ch == '1' || ch == '4')
        mode = 3;
      else
        return ch;
    }
    else if (mode == 2)
    {                           /* Cursor key */
      if (ch >= 'A' && ch <= 'D')
        return KEY_UP + (ch - 'A');
      else if (ch >= '1' && ch <= '6')
        mode = 3;
      else
        return ch;
    }
    else if (mode == 3)
    {                           /* Ins Del Home End PgUp PgDn */
      if (ch == '~')
        return KEY_HOME + (last - '1');
      else
        return ch;
    }
    last = ch;
  }
}
#endif                          /* TRAP_ESC */

==> FHLBBS-5.3.tgz: maple/io.c <==
/*-------------------------------------------------------*/
/* io.c         ( NTHU CS MapleBBS Ver 2.36 )            */
/*-------------------------------------------------------*/
/* target : basic console/screen/keyboard I/O routines   */
/* create : 95/02/28                                     */
/* update : 95/12/15                                     */
/*-------------------------------------------------------*/

#include "bbs.h"

#ifdef AIX
#include <sys/select.h>
#endif

#ifdef  Linux
#define OBUFSIZE  (2048)
#define IBUFSIZE  (128)
#else
#define OBUFSIZE  (4096)
#define IBUFSIZE  (256)
#endif

#define INPUT_ACTIVE    0
#define INPUT_IDLE      1

static char outbuf[OBUFSIZE];
static int obufsize = 0;

static char inbuf[IBUFSIZE];
static int ibufsize = 0;
static int icurrchar = 0;

static int i_mode = INPUT_ACTIVE;

extern int dumb_term;


/* ----------------------------------------------------- */
/* wܰʺAݪO                                      */
/* ----------------------------------------------------- */


#ifdef  HAVE_MOVIE
#define STAY_TIMEOUT    (30*60)
extern void movie();

static void
hit_alarm_clock()
{
  static int stay_time = 0;
  static int idle_time = 0;

  if (currutmp->pid != currpid)
  {
    setup_utmp(XMODE);          /* stm shm */
  }

  if ((i_mode == INPUT_IDLE) && (!HAS_PERM(PERM_NOTIMEOUT)))
  {
    idle_time += MOVIE_INT;
    if (idle_time > IDLE_TIMEOUT)
    {
      clear();
      fprintf(stderr, "WLmɶIBooting...\n");
      abort_bbs();
    }
  }
  else
  {
    idle_time = 0;
    i_mode = INPUT_IDLE;
  }

  if (currstat && currstat < CLASS)
    movie(0);
  alarm(MOVIE_INT);
 

  stay_time += MOVIE_INT;
  if (chkmail(0) && stay_time > 10 * 60 || stay_time > STAY_TIMEOUT)
  {
    /* bo̴ user 𮧤@U */
    char *msg[8] = {"y, ||, ܤf....",
                    "zwgFbӤpo! _Ӭʤ@Ua!",
                    "ŶܡHХemailtalk߻ݭnߪH",
                    "AUh||[`O~~~",
                    "SnƱѤFBzO?",
                    "ɥiHQQSHݭnߡH",
		    "OѤFnWWZҡAܳܤAHKoH֪εhI",
                    "OѤFq᭱]OӬͥͪHI"};
/*
woju
*/
    time_t now = time(0);

    char buf[100];
    int i = rand() % 8;

    sprintf(currmsg, "[1;33;41m[%s] %s[m", Cdate(&now), 
       chkmail(0) ? "Hc٦SݹLH@!" : msg[i]);
    kill(currpid, SIGUSR2);
    stay_time = 0;
  }
  if (idle_time > IDLE_TIMEOUT - 60)
  {
    sprintf(currmsg, "[1;5;37;41mĵiGzwmL[A"
                       "YL^AtΧYNII[m");
    kill(currpid, SIGUSR2);
  }
}


void
init_alarm()
{
  alarm(0);
  signal(SIGALRM, hit_alarm_clock);
  alarm(MOVIE_INT);
}

#else                           /* HAVE_MOVIE */

static void
hit_alarm_clock()
{
  if (HAS_PERM(PERM_NOTIMEOUT))
    return;
  if (i_mode == INPUT_IDLE)
  {
    clear();
    fprintf(stderr, "WLmɶIBooting...\n");
    kill(currpid, SIGHUP);
  }
  i_mode = INPUT_IDLE;
  alarm(IDLE_TIMEOUT);
}


void
init_alarm()
{
  signal(SIGALRM, hit_alarm_clock);
  alarm(IDLE_TIMEOUT);
}
#endif                          /* HAVE_MOVIE */


/* ----------------------------------------------------- */
/* output routines                                       */
/* ----------------------------------------------------- */


void
oflush()
{
  if (obufsize)
  {
    write(1, outbuf, obufsize);
    obufsize = 0;
  }
}


void
output(s, len)
  char *s;
{
  /* Invalid if len >= OBUFSIZE */

  if (obufsize + len > OBUFSIZE)
  {
    write(1, outbuf, obufsize);
    obufsize = 0;
  }
  memcpy(outbuf + obufsize, s, len);
  obufsize += len;
}


void
ochar(c)
{
  if (obufsize > OBUFSIZE - 1)
  {
    write(1, outbuf, obufsize);
    obufsize = 0;
  }
  outbuf[obufsize++] = c;
}


/* ----------------------------------------------------- */
/* input routines                                        */
/* ----------------------------------------------------- */


static int i_newfd = 0;
static struct timeval i_to, *i_top = NULL;
static int (*flushf) () = NULL;


void
add_io(fd, timeout)
  int fd;
  int timeout;
{
  i_newfd = fd;
  if (timeout)
  {
    i_to.tv_sec = timeout;
    i_to.tv_usec = 0;
    i_top = &i_to;
  }
  else
    i_top = NULL;
}


void
add_flush(flushfunc)
  int (*flushfunc) ();
{
  flushf = flushfunc;
}


int
num_in_buf()
{
  return icurrchar - ibufsize;
}


int
igetch()
{
  int ch;

  for (;;)
  {
    if (ibufsize == icurrchar)
    {
      fd_set readfds;
      struct timeval to;

      to.tv_sec = to.tv_usec = 0;
      FD_ZERO(&readfds);
      FD_SET(0, &readfds);
      if (i_newfd)
        FD_SET(i_newfd, &readfds);
      if ((ch = select(FD_SETSIZE, &readfds, NULL, NULL, &to)) <= 0)
      {
        if (flushf)
          (*flushf) ();

        if (dumb_term)
          oflush();
        else
          refresh();

        FD_ZERO(&readfds);
        FD_SET(0, &readfds);
        if (i_newfd)
          FD_SET(i_newfd, &readfds);

        while ((ch = select(FD_SETSIZE, &readfds, NULL, NULL, i_top)) < 0)
        {
          if (errno == EINTR)
            continue;
          else
          {
            perror("select");
            return -1;
          }
        }
        if (ch == 0)
          return I_TIMEOUT;
      }
      if (i_newfd && FD_ISSET(i_newfd, &readfds))
        return I_OTHERDATA;

      while ((ibufsize = read(0, inbuf, IBUFSIZE)) <= 0)
      {
        if (ibufsize == 0)
          longjmp(byebye, -1);
        if (ibufsize < 0 && errno != EINTR)
          longjmp(byebye, -1);
      }
      icurrchar = 0;
    }

    i_mode = INPUT_ACTIVE;
    ch = inbuf[icurrchar++];
    if (ch != Ctrl('L'))
      return (ch);

    redoscr();
  }
}

char* phone_char(char c)
{
   switch (c) {
          case '1':
             return "t";
          case 'q':
             return "u";
          case 'a':
             return "v";
          case 'z':
             return "w";
          case '2':
             return "x";
          case 'w':
             return "y";
          case 's':
             return "z";
          case 'x':
             return "{";
          case 'e':
             return "|";
          case 'd':
             return "}";
          case 'c':
             return "~";
          case 'r':
             return "";
          case 'f':
             return "";
          case 'v':
             return "";
          case '5':
             return "";
          case 't':
             return "";
          case 'g':
             return "";
          case 'b':
             return "";
          case 'y':
             return "";
          case 'h':
             return "";
          case 'n':
             return "";
          case 'u':
             return "";
          case 'j':
             return "";
          case 'm':
             return "";
          case '8':
             return "";
             break;
          case 'i':
             return "";
          case 'k':
             return "";
          case ',':
             return "";
          case '9':
             return "";
          case 'o':
             return "";
          case 'l':
             return "";
          case '.':
             return "";
          case '0':
             return "";
          case 'p':
             return "";
          case ';':
             return "";
          case '/':
             return "";
          case '-':
             return "";
          case '6':
             return "";
          case '3':
             return "";
          case '4':
             return "";
          case '7':
             return "";
   }
  return 0;
}


int
getdata(line, col, prompt, buf, len, echo)
  int line, col;
  char *prompt, *buf;
  int len, echo;
{
  register int ch;
  int clen;
  int x, y;
  extern unsigned char scr_cols;
#define MAXLASTCMD 6
  static char lastcmd[MAXLASTCMD][80];

  if (prompt)
  {
    move(line, col);
    clrtoeol();
    outs(prompt);
  }
  else
    clrtoeol();

  if (dumb_term || !echo)
  {
    len--;
    clen = 0;
    while ((ch = igetch()) != '\r')
    {
      if (ch == '\n')
        break;
      if (ch == '\177' || ch == Ctrl('H'))
      {
        if (!clen)
        {
          bell();
          continue;
        }
        clen--;
        if (echo)
        {
          ochar(Ctrl('H'));
          ochar(' ');
          ochar(Ctrl('H'));
        }
        continue;
      }

#ifdef BIT8
      if (!isprint2(ch))
#else
      if (!isprint(ch))
#endif

      {
        if (echo)
          bell();
        continue;
      }
      if (clen >= len)
      {
        if (echo)
          bell();
        continue;
      }
      buf[clen++] = ch;
      if (echo)
        ochar(ch);
    }
    buf[clen] = '\0';
    outc('\n');
    oflush();
  }
  else
  {
   int cmdpos = -1;
   int currchar = 0;

    getyx(&y, &x);
    standout();
    for (clen = len--; clen; clen--)
      outc(' ');
    standend();
    memset(buf, 0, len);

    while (move(y, x + currchar), (ch = igetkey()) != '\r')
    {
/*
woju
*/
       switch (ch) {

       case KEY_DOWN:
       case Ctrl('N'):
          cmdpos += MAXLASTCMD - 2;
       case Ctrl('P'):
       case KEY_UP: {
          int i;

          cmdpos++;
          cmdpos %= MAXLASTCMD;
          strncpy(buf, lastcmd[cmdpos], len);

          move(y, x);                   /* clrtoeof */
          for (i = 0; i <= clen; i++)
             outc(' ');
          move(y, x);

          outs(buf);
          clen = currchar = strlen(buf);
          continue;
       }
       case KEY_LEFT:
          if (currchar)
             --currchar;
          continue;
       case KEY_RIGHT:
          if (buf[currchar])
             ++currchar;
          continue;
       }

      if (ch == '\n' || ch == '\r')
         break;
/*
woju
*/
      if (currutmp->msgs[0].last_pid && ch == Ctrl('R') &&
          !(currutmp->mode == 0 &&
            (currutmp->chatid[0] == 2 || currutmp->chatid[0] == 3))) {
         extern screenline* big_picture;
         screenline* screen0 = calloc(t_lines, sizeof(screenline));

         memcpy(screen0, big_picture, t_lines * sizeof(screenline));

         show_last_call_in();
         //my_write(currutmp->msgs[0].last_pid, "Te^hG");
         memcpy(big_picture, screen0, t_lines * sizeof(screenline));
         free(screen0);
         redoscr();
         continue;
      }                                                                       
      if (ch == Ctrl('U') && currstat != IDLE  &&
          !(currutmp->mode == 0 &&
            (currutmp->chatid[0] == 2 || currutmp->chatid[0] == 3))) {
         extern screenline* big_picture;
         screenline* screen0 = calloc(t_lines, sizeof(screenline));

         memcpy(screen0, big_picture, t_lines * sizeof(screenline));

         t_users();
         memcpy(big_picture, screen0, t_lines * sizeof(screenline));
         free(screen0);
         redoscr();
         continue;
      }                                                                       
      if (ch == Ctrl('I') && currstat != IDLE &&
          !(currutmp->mode == 0 &&
            (currutmp->chatid[0] == 2 || currutmp->chatid[0] == 3))) {
         extern screenline* big_picture;
         screenline* screen0 = calloc(t_lines, sizeof(screenline));

         memcpy(screen0, big_picture, t_lines * sizeof(screenline));

         t_idle();
         memcpy(big_picture, screen0, t_lines * sizeof(screenline));
         free(screen0);
         redoscr();
         continue;
      }                                                                       
      if (ch == '\177' || ch == Ctrl('H'))
      {
        if (currchar) {
           int i;

           currchar--;
           clen--;
           for (i = currchar; i <= clen; i++)
              buf[i] = buf[i + 1];
           move(y, x + clen);
           outc(' ');
           move(y, x);
           outs(buf);
        }
/*
        if (clen)
        {
          clen--;
          move(y, x + clen);
          outc(' ');
          move(y, x + clen);
        }
*/
        continue;
      }
      if (ch == Ctrl('Y')) {
         int i;

         buf[0] = 0;
         currchar = 0;
         move(y, x);                    /* clrtoeof */
         for (i = 0; i < clen; i++)
            outc(' ');
         clen = 0;
         continue;
      }
      if (ch == Ctrl('C')) {
         buf[0] = 0;
         break;
      }
      if (ch == Ctrl('D')) {
        if (buf[currchar]) {
           int i;

           clen--;
           for (i = currchar; i <= clen; i++)
              buf[i] = buf[i + 1];
           move(y, x + clen);
           outc(' ');
           move(y, x);
           outs(buf);
        }
        continue;
      }
      if (ch == Ctrl('K')) {
         int i;

         buf[currchar] = 0;
         move(y, x + currchar);
         for (i = currchar; i < clen; i++)
            outc(' ');
         clen = currchar;
         continue;
      }
      if (ch == Ctrl('A')) {
         currchar = 0;
         continue;
      }
      if (ch == Ctrl('E')) {
         currchar = clen;
         continue;
      }


#ifdef BIT8
      if (!isprint2(ch))
#else
      if (!isprint(ch))
#endif

      {
        continue;
      }
      if (clen >= len || x + clen >= scr_cols)
      {
        continue;
      }
/*
woju
*/
      if (buf[currchar]) {               /* insert */
         int i;

         for (i = currchar; buf[i] && i < len && i < 80; i++)
            ;
         buf[i+1] = 0;
         for (; i > currchar; i--)
            buf[i] = buf[i - 1];
      }
      else                              /* append */
         buf[currchar + 1] = '\0';
      buf[currchar] = ch;
      move(y, x + currchar);
      outs(buf + currchar);
      currchar++;
      clen++;
/*
      move(y, x + clen);
      buf[clen++] = ch;
      outc(echo ? ch : '*');
*/
    }
    buf[clen] = '\0';
    if (clen > 1) {
       for (cmdpos = MAXLASTCMD - 1; cmdpos; cmdpos--)
          strcpy(lastcmd[cmdpos], lastcmd[cmdpos - 1]);
       strncpy(lastcmd[0], buf, len);
    }
    if (echo)
      outc('\n');
    refresh();
  }
  if ((echo == LCECHO) && ((ch = buf[0]) >= 'A') && (ch <= 'Z'))
    buf[0] = ch | 32;
  return clen;
}


/*
woju
*/
#define TRAP_ESC

#ifdef  TRAP_ESC
int KEY_ESC_arg;

int
igetkey()
{
  int mode;
  int ch, last;

  mode = last = 0;
  while (1)
  {
    ch = igetch();
    if (mode == 0)
    {
      if (ch == KEY_ESC)
        mode = 1;
      else
        return ch;              /* Normal Key */
    }
    else if (mode == 1)
    {                           /* Escape sequence */
      if (ch == '[' || ch == 'O')
        mode = 2;
      else if (ch == '1' || ch == '4')
        mode = 3;
      else
      {
        KEY_ESC_arg = ch;
        return KEY_ESC;
      }
    }
    else if (mode == 2)
    {                           /* Cursor key */
      if (ch >= 'A' && ch <= 'D')
        return KEY_UP + (ch - 'A');
      else if (ch >= '1' && ch <= '6')
        mode = 3;
      else
        return ch;
    }
    else if (mode == 3)
    {                           /* Ins Del Home End PgUp PgDn */
      if (ch == '~')
        return KEY_HOME + (last - '1');
      else
        return ch;
    }
    last = ch;
  }
}

#else                           /* TRAP_ESC */

int
igetkey(void)
{
  int mode;
  int ch, last;

  mode = last = 0;
  while (1)
  {
    ch = igetch();
    if (ch == KEY_ESC)
      mode = 1;
    else if (mode == 0)         /* Normal Key */
      return ch;
    else if (mode == 1)
    {                           /* Escape sequence */
      if (ch == '[' || ch == 'O')
        mode = 2;
      else if (ch == '1' || ch == '4')
        mode = 3;
      else
        return ch;
    }
    else if (mode == 2)
    {                           /* Cursor key */
      if (ch >= 'A' && ch <= 'D')
        return KEY_UP + (ch - 'A');
      else if (ch >= '1' && ch <= '6')
        mode = 3;
      else
        return ch;
    }
    else if (mode == 3)
    {                           /* Ins Del Home End PgUp PgDn */
      if (ch == '~')
        return KEY_HOME + (last - '1');
      else
        return ch;
    }
    last = ch;
  }
}
#endif                          /* TRAP_ESC */
